/* a06_mtrx.c
 * Matrixtyp a06_matrix
 * erstellt: Fre Nov  5 15:04:28 CET 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include "matrix.h"
#include "namen.h"
#include "a06_mtrx.h"
#include "nmmatrix.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>


/* externes Klassenobjekt einbinden */

extern matrix nm_matrix;


/* Deklaration der Objekt-Methoden */

matrix * a06_matrix_laden(FILE *datei, void *vater);
int a06_matrix_speichern(matrix *m, FILE *datei);
matrix * a06_matrix_kreieren(void *schablone, void *vater);
int a06_matrix_freigeben(matrix *m, void *vater);
double a06_matrix_amIndex(matrix *m, int zeile, int spalte);
void a06_matrix_setzen(matrix *m, int zeile, int spalte, double d);
int a06_matrix_breite(matrix *m);
int a06_matrix_hoehe(matrix *m);
char *a06_matrix_name(matrix *m);
matrix * a06_matrix_imVektor(matrix *m, int vektor);


/* Globales Klassenobjekt */

matrixFunktionen a06__matrixFunktionen = {
    matrix_ausgeben,
    a06_matrix_speichern,
    a06_matrix_laden,
    a06_matrix_freigeben,
    a06_matrix_kreieren,
    a06_matrix_amIndex,
    a06_matrix_setzen,
    a06_matrix_breite,
    a06_matrix_hoehe,
    a06_matrix_name,
    matrix_klasse,
    nmmatrix_produkt,
    a06_matrix_imVektor
};

matrix a06__matrix = {          /* das Klassenobjekt zur Klasse a06__MATRIX */ 
    &a06__matrixFunktionen,     
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* private Methoden */

static a06_matrixdaten *get_a06_matrixdaten(matrix *m);
static matrix * a06_matrix_pkreieren(void *schablone, void *vater);
static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum);


/* ------------------------------------------------------------------------ 
 * a06_-Matrix Implementation
 */ 


matrix * a06_matrix_laden(FILE *datei, void *vater)
{
  a06_matrixschablone schablone;
  int i;
  matrix *m;
  a06_matrixdaten *pm;

  schablone.name = matrix_kopfzeile(datei, S_A06__MATRIX);

  /* Abmasse und a06_breite der Matrix feststellen */

  lese_datum(datei, "%d", "n =", &(schablone.n));

  /* Rohmatrix ohne Diagonal-Vektoren kreieren */

  m = a06_matrix_pkreieren(&schablone, vater);

  /* Eintraege aus der Datei lesen */

  pm = get_a06_matrixdaten(m);

  for (i = 0; i < 3; i++)
    pm->abc[i] = nm_matrix.aktion->laden(datei, m);

  return m;
}


static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum)
{
  char eingabe[BUFSIZ];
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Zeile %s nicht gefunden - EOF.\n", prefix);
    exit(1);
  }
  if (strncmp(eingabe, prefix,
               strlen( prefix))!=0) {
    fprintf(stderr,"%s... erwartet!\n", prefix);
    exit(1);
  }
  sscanf(&(eingabe[strlen(prefix)]), format, pdatum);
                               /* erstmal ohne Fehlerkontrolle */
  return 0;
}


int a06_matrix_speichern(matrix *m, FILE *datei)
{
  a06_matrixdaten *pm;
  int i;
  pm = get_a06_matrixdaten(m);
  fprintf(datei,S_A06__MATRIX ": %s\n", pm->name);
  fprintf(datei,"n = %d\n", pm->n);
  for (i = 0; i < 3; i++)
    pm->abc[i]->aktion->speichern(pm->abc[i], datei);
  return 0;
}


static matrix * a06_matrix_pkreieren(void *schablone, void *vater)
{
  a06_matrixschablone *s = (a06_matrixschablone *) schablone;
  matrix *m;
  a06_matrixdaten *pm;
  char *name;
  if (!(m = (matrix *)malloc(sizeof(matrix)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue Matrix (Teil 1)!\n");
    exit(1);          
  }
  m->aktion = a06__matrix.aktion;    /* moegliche Aktionen festlegen */
  if (!(m->daten = (a06_matrixdaten *)malloc(sizeof(a06_matrixdaten)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue Matrix (Teil 2)!\n");
    exit(1);          
  }
  pm = get_a06_matrixdaten(m);

  /* Name der Matrix festlegen */

  name = (s && s->name)?s->name:"unbenannt";
  if (!(pm->name = (char *)malloc( sizeof(char) * (1+strlen(name))))) {
    fprintf(stderr, "Nicht genug Speicher fuer " S_A06__MATRIX " %s.\n",
                    name);
    exit(1);
  }
  strcpy(pm->name, name);

  pm->n = s?s->n:5;  /* Groesse der Matrix ggf. aus Schablone, sonst 5x5 */

  if (!(pm->abc = (matrix **)malloc( 3 * sizeof(matrix *) ))) {
    fprintf(stderr,"Nicht genug Speicher fuer Matrixeintraege!\n");
    exit(1);          
  }
  return m;
}


matrix * a06_matrix_kreieren(void *schablone, void *vater)
{
  a06_matrixschablone *s = (a06_matrixschablone *) schablone;
  matrix *m;
  a06_matrixdaten *pm;
  nmmatrixschablone schablone2; 
  
  /* Rohmatrix ohne Diagonal-Vektoren kreieren */

  m = a06_matrix_pkreieren(schablone, vater);
  pm = get_a06_matrixdaten(m);

  /* Diagonal-Vektor alloziieren */

  schablone2.n = pm->n;  /* Vektor als Spaltenvektor speichern, also */
  schablone2.m = 1;      /* eine n x 1 Matrix */
  schablone2.init = s->init;
  schablone2.name = "Hauptdiagonal-Vektor";
  pm->abc[0] = nm_matrix.aktion->kreieren(&schablone2, m); 

  /* Nebendiagonal-Vektor alloziieren */
 
  schablone2.n = pm->n - 1;  /* Vektor als Spaltenvektor speichern, also */
  schablone2.m = 1;          /* eine (n-1) x 1 Matrix */
  schablone2.name = "Nebendiagonal-Vektor";
  pm->abc[1] = nm_matrix.aktion->kreieren(&schablone2, m); 

  /* Element c alloziieren */

  schablone2.n = 1;          /* c als eine 1 x 1 Matrix speichern */
  schablone2.m = 1;          
  schablone2.name = "Eckelement c";
  pm->abc[2] = nm_matrix.aktion->kreieren(&schablone2, m); 

  return m;
}

int a06_matrix_freigeben(matrix *m, void *vater) 
{
  int i; 
  a06_matrixdaten *pm;
  pm = get_a06_matrixdaten(m);
  for(i=0; i < 3; i++) 
    pm->abc[i]->aktion->freigeben(pm->abc[i],m);
  free(pm->name);
  free(pm->abc);     
  free(pm);
  free(m);
  return 0;              /* immer Erfolg */
}

static a06_matrixdaten *get_a06_matrixdaten(matrix *m) 
{
  if ((!(m->daten))) /* sicher ist sicher ... */
    exit(1);                        /* Fehlermeldung fehlt noch */
  return m->daten;                  /* Daten der Matrix zurueckgeben */
}


#define a06_matrix_pruefeIndex(pm, zeile, spalte)  \
  if ((zeile < 1) ||  \
      (spalte < 1) || \
      (zeile > pm->n) || \
      (spalte > pm->n)) { \
   fprintf(stderr,S_A06__MATRIX "Ungueltiger Index fuer Matrix: (%i,%i)\n", \
                  zeile, spalte); \
   exit(1); \
  } 


double a06_matrix_amIndex(matrix *m, int zeile, int spalte)
{
  a06_matrixdaten *pm;
  pm = get_a06_matrixdaten(m);
  a06_matrix_pruefeIndex(pm, zeile, spalte); 

  if(zeile==spalte) {
    return pm->abc[0]->aktion      /*  i-te Komponente im Diagonalvektor */
      ->amIndex(pm->abc[0], zeile, 1);  /* zurueckgeben */
  }
  switch (zeile - spalte) {
    case -1: return pm->abc[1]->aktion          /* obere Nebendiagonale */
               ->amIndex(pm->abc[1], zeile, 1);
    case  1: return pm->abc[1]->aktion          /* untere Nebendiagonale */
               ->amIndex(pm->abc[1], spalte, 1);
  }
  if (abs(zeile-spalte) == (pm->n-1)) 
    return pm->abc[2]->aktion          /* Eckelement */
      ->amIndex(pm->abc[2], 1, 1);
  else
    return 0.0;
}

void a06_matrix_setzen(matrix *m, int zeile, int spalte, double d)
{
  a06_matrixdaten *pm;
  pm = get_a06_matrixdaten(m);
  a06_matrix_pruefeIndex(pm, zeile, spalte); 

  if(zeile==spalte) {
    pm->abc[0]->aktion             /*  i-te Komponente im Diagonalvektor */
      ->setzen(pm->abc[0], zeile, 1, d);  /* zurueckgeben */
    return;
  }
  switch (zeile - spalte) {
    case -1: pm->abc[1]->aktion          /* obere Nebendiagonale */
               ->setzen(pm->abc[1], zeile, 1, d);
             return;
             break;
    case  1: pm->abc[1]->aktion          /* untere Nebendiagonale */
               ->setzen(pm->abc[1], spalte, 1, d);
             return;
             break;
  }
  if (abs(zeile-spalte) == (pm->n-1)) {
    pm->abc[2]->aktion                   /* Eckelement */
      ->setzen(pm->abc[2], 1, 1, d);
    return;
  }else{
    if (fabs(d) >= EPSILON) {
      fprintf(stderr, "Fehler: es wurde versucht, in der " S_A06__MATRIX 
                      " \'%s\' an der Stelle (%i,%i) einen Wert ungleich "
                      "null (%g) zu setzen.\n", pm->name, zeile, spalte, d);
      exit(1);
    }
  }
}

int a06_matrix_breite(matrix *m)
{
  a06_matrixdaten *pm;
  pm = get_a06_matrixdaten(m);
  return pm->n;
}

int a06_matrix_hoehe(matrix *m)
{
  a06_matrixdaten *pm;
  pm = get_a06_matrixdaten(m);
  return pm->n;
}

char *a06_matrix_name(matrix *m)
{
  a06_matrixdaten *pm;
  if (!m->daten)
    return S_A06__MATRIX;
  pm = get_a06_matrixdaten(m);
  return pm->name;
}

matrix * a06_matrix_imVektor(matrix *m, int vektorNr)
{
  a06_matrixdaten *pm;
  pm = get_a06_matrixdaten(m);
  if (vektorNr >= 3) {
    fprintf(stderr,"Ungueltiger Diagonal-Vektor fuer a06_matrix %s: (%i)\n",
                   pm->name, vektorNr);
    exit(1); 
  }
  return pm->abc[vektorNr];
}
