/* matrix.c
 * nuetzliche Funktionen fuer allgemeine Matrixen 
 * erstellt: Fre Nov  5 17:46:37 CET 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */


#include "matrix.h"
#include "namen.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* externe Klassenobjekte einbinden */

extern matrix nm_matrix;
extern matrix band_matrix;
extern matrix a06__matrix;


/* Deklaration der Objekt-Methoden */

matrix * matrix_laden(FILE *datei, void *vater);


/* Globales Klassenobjekt */

matrixFunktionen allg_matrixFunktionen = {
    NULL,
    NULL,
    matrix_laden
};

matrix allg_matrix = {        /* das Klassenobjekt zur Klasse ALLG_MATRIX */
    &allg_matrixFunktionen,
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* ------------------------------------------------------------------------
 * Implementation
 */

/*
 * matrix_laden
 * stellt den Typ der Matrix fest und veranlasst dann das
 * Laden der Matrix
 */

matrix * matrix_laden(FILE *datei, void *vater)
{
  char typ[80];
  typ[0] = '\0';                      /* leerer String */
  matrix_kopfzeile(datei, typ);       /* Matrixtyp ermitteln */

  if (strcmp(typ, S_NM_MATRIX)==0) 
    return nm_matrix.aktion->laden(datei, vater);

  if (strcmp(typ, S_BAND_MATRIX)==0) 
    return band_matrix.aktion->laden(datei, vater);

  if (strcmp(typ, S_A06__MATRIX)==0) 
    return a06__matrix.aktion->laden(datei, vater);

  fprintf(stderr, "Matrix %s vom Matrixtyp %s kann nicht eingelesen "
                  "werden, da dieser Typ nicht im Klassenkluster "
                  "registriert.n", matrix_kopfzeile(datei, typ), typ);
  exit(1);
}


/* 
 * stellt die Matrix tabellarisch dar  
 */

int matrix_ausgeben(matrix *a, FILE *datei)
{
  int i,j;
  int n, m;
  double wert;
  n = a->aktion->hoehe(a);
  m = a->aktion->breite(a);
  fprintf(datei,"Matrix \'%s\':\n", a->aktion->name(a));
  for (i=1; i<=n; i++) {
    for (j=1; j<=m; j++) {
      wert = a->aktion->amIndex(a, i, j);
      if ((wert == 0) 
          || (fabs(1.0/wert) < 10000.0)) {
        if (fabs(wert) < 10.0) 
          fprintf(datei,"%9.6f%c", wert, j<m?' ':'\n');
        else
          if (fabs(wert) < 10000.0) 
            fprintf(datei,"%9.3f%c", wert, j<m?' ':'\n');
          else
            fprintf(datei,"%9.3g%c", wert, j<m?' ':'\n');
      }else{
        fprintf(datei,"%9.3g%c", wert, j<m?' ':'\n');
      }
    }
  }
  return 1;  /* Erfolg */
}


/* matrix_kopfzeile
 * liest die Kopfzeile eine Matrix ein, ueberprueft den Typ der Matrix
 * und gibt den Namen der Matrix als lokalen String zureuck, den sich
 * der Aufrufer zur weiteren Verwendung kopieren muss.
 * Wird ein leerer String als matrixtyp uebergeben, so wird der Typ er-
 * mittelt und in matrixtyp abgelegt. Ein nachfolgender Aufruf dieser
 * Funktion bearbeitet dann keine neue Zeile sondern liefert den Namen
 * der Matrix, deren Typ zuvor bestimmt wurde.
 */

static int kopfzeile_halb_gelesen = 0;
static char eingabe[BUFSIZ];
static char unbenannt[] = "unbenannt";
static char *name;

char *matrix_kopfzeile(FILE *datei, char *matrixtyp) 
{
  char *p = NULL;

  if (!kopfzeile_halb_gelesen) {             /* Zeile noch nicht gelesen? */

    if (!fgets(eingabe,BUFSIZ,datei)) {          /* Zeile einlesen */
      fprintf(stderr,"Keine Daten fuer neue Matrix gefunden - EOF.\n");
      exit(1);
    }

    /* Matrixtyp feststellen */

    if (! (p = strchr(eingabe,':'))) {            /* Trennstelle suchen */
      fprintf(stderr,"Fehler in der Eingabe: "
                     "<Matrixtyp>:[Matrixname] erwartet.\n");
      exit(1);
    }
    *p = '\0';                                   /* String-Ende setzen */
  }

  if (*matrixtyp) {                           /* Matrixtyp vorgegeben? */
    if (strcmp(eingabe, matrixtyp)!=0) {      /* Ja -> ueberpruefen    */
      fprintf(stderr,"Dies ist keine %s Matrix!\n", matrixtyp);
      exit(1);
    }
    if (kopfzeile_halb_gelesen) {
      kopfzeile_halb_gelesen = 0;
      return name;
    }
  }else{                                      /* Nein -> in den bereitge- */
    strcpy(matrixtyp, eingabe);               /* stellten String kopieren */
    if (kopfzeile_halb_gelesen) {
      fprintf(stderr,"Fehler: Kopfzeile wurde bereits gelesen\n");
      exit(1);
    }
    kopfzeile_halb_gelesen = 1;
  }

  if (!p) {                                   /* Dieser Fehler kann nicht */
    fprintf(stderr,"Fehler in matrix_kopfzeile\n");          /* auftreten */
    exit(1);
  } 

  /* Name der Matrix extrahieren */
  
  p++;                             /* hinter den Ex-':' schauen */
  while (*p && isspace(*p)) p++;   /* Anfang finden         */
  if (p[strlen(p)-1] == '\n')      /* ggf. abschliessendes NewLine tilgen */
    p[strlen(p)-1] = '\0';
  if (!*p) {                       /* Falls keinen gefunden */
    name = unbenannt;              /* default setzen        */
  }else{
    name = p;
  }
  return name;                     /* zurueckgeben */
}


/* matrix_klasse
 * liefert Art der Matrix
 */


extern matrix nm_matrix;
extern matrix band_matrix;
extern matrix a06__matrix;

matrix * matrix_klasse(matrix *m)
{
  if (m->aktion == nm_matrix.aktion)
    return &nm_matrix;
  if (m->aktion == band_matrix.aktion)
    return &band_matrix;
  if (m->aktion == a06__matrix.aktion)
    return &a06__matrix;
  fprintf(stderr, "Fehler: undefiniertes Objekt %s\n", m->aktion->name(m));
  exit(1);
}

